#include "svd.h"
#include <Accelerate/Accelerate.h>
#include <stdio.h>
#include <time.h>

/* Prints an M x N matrix 'mat' (column-major order) */
void printmat(feature *mat, long int M, long int N) {
  long int i, j;
  for(i = 0; i < M; i++) {
    for(j = 0; j < N; j++)
      printf("%f ", mat[i+j*M]);
    printf("\n");
  }
}
void iprintmat(int *mat, long int M, long int N) {
  long int i, j;
  for(i = 0; i < M; i++) {
    for(j = 0; j < N; j++)
      printf("%d ", mat[i+j*M]);
    printf("\n");
  }
}

/*
 *  Produces an M x N matrix (column-major order) filled
 *  with values on the interval (0.0,1.0)
 */
feature* randmat(long int M, long int N) {
  feature *mat = malloc(M*N*sizeof(feature));
  long int i, j;
  if(mat == NULL) {
    perror("randmat failed to allocate!");
  }
  for(i = 0; i < M; i++)
    for(j = 0; j < N; j++)
      mat[i+j*M] = rand_0_1();
  return mat;
}

feature* krank_randmat(long int M, long int N, long int K) {
  feature *A, *B, *C, tmp;
  int i, j, k;

  A = randmat(M, K);
  B = randmat(K, N);
  C = malloc(sizeof(feature)*M*N);

  for(i = 0; i < M; i++) {
    for(j = 0; j < N; j++) {
      tmp = 0.0;
      for(k = 0; k < K; k++) {
        tmp += A[i+k*M]*B[k+j*K];
      }
      C[i+j*M] = tmp;
    }
  }

  free(A); free(B);
  return C;
}

/**
 * Inverts an M x N matrix 'mat' (column-major order) 
 * 0 on success
 * Otherwise, failure 
 */
inline int invmat(feature *mat, long int M, long int N) {
  long int *ipiv, lwork, info;
  feature *work;

  ipiv = malloc(sizeof(long int)*N);
  lwork = -1;
  work = malloc(sizeof(feature));

  GETRI(&N, mat, &M, ipiv, work, &lwork, &info);
  if(info != 0) {
    //perror("LWORK query failed!\n");
    return 1;
  }

  GETRF(&M, &N, mat, &M, ipiv, &info);
  if(info != 0) {
    //perror("DGETRF failed!\n");
    return 1;
  }

  lwork = work[0]*N;
  free(work);
  work = malloc(sizeof(feature)*lwork);

  GETRI(&N, mat, &M, ipiv, work, &lwork, &info);
  if(info != 0) {
    //perror("DGETRI failed!\n");
    return 1;
  }

  free(ipiv);
  free(work);

  return 0;
}

void invtest(int N_ITER, int N_FEAT) {
  long int i, j, iterations, M, N;
  int failures;
  feature *A;
  time_t start, stop;
  struct tm *timeinfo;

  time(&start);
  timeinfo = localtime(&start);
  printf("Starting @ %s\n", asctime(timeinfo));

  for(iterations = 0; iterations < N_ITER; iterations++) {
    M = N_FEAT;
    N = M;
    A = malloc(sizeof(feature)*M*N);
    for(j = 0; j < N; j++)
      for(i = 0; i < M; i++)
        A[i+M*j] = rand() % 5 + 1;
    failures += invmat(A,M,N);
    free(A);
  }

  time(&stop);
  timeinfo = localtime(&stop);
  printf("Ending   @ %s\n", asctime(timeinfo));
  printf("%f seconds elapsed\n", difftime(stop, start));
  printf("%d iterations\n %d failures\n", iterations, failures);
}
