/* svd.h - Peter Nelson
 *
 * Headers for doing SVD on the netflix dataset
 * 
 * Also: Includes utility functions for dealing with matrices stored in 
 * column-major order (as in Fortran and LAPACK, but not as in normal C)
 */
#ifndef SVD_H
#define SVD_H

#include <stdio.h>
#include "../../netflixcore.h"

/* Double or single precision */
#ifdef DOUBLE
typedef double feature;
#define GETRI dgetri_
#define GETRF dgetrf_
#else
typedef float feature;
#define GETRI sgetri_
#define GETRF sgetrf_
#endif

/* Utility stuff */

/* Matrix Utility Functions */
void printmat(feature *mat, long int M, long int N);
void iprintmat(int *mat, long int M, long int N);
feature* randmat(long int M, long int N);
feature* krank_randmat(long int M, long int N, long int K);
inline int invmat(feature *mat, long int M, long int N);

/* Top-n list functionality */
struct TopNNode {
  double rating;         /* Rating (stars) */
  int mid;               /* Movie ID */
  int offset;            /* Used for finding names */
  struct TopNNode *next;
};
typedef struct TopNNode TopNNode;
TopNNode* topn_addrating(TopNNode *first, double rating, int mid, int n);
void topn_freeall(TopNNode *first, int n);
void topn_trim(TopNNode *first, int n);
int* topn_findoffsets(TopNNode *first, int n);

/* SVD Stuff */

/* Number of features for doing SVD */
#define NFEATURES 60

/* Holds the info for doing svd (per user and movie) */
typedef struct {
  feature* features; /* The (low) dimensional vector associated with each user/movie */
  int nvisible;      /* The number of visible ratings by this user/for this movie */
  double err;        /* The current error associated with the feature vector */
  double bias;       /* The bias corrector for this user/movie */
} svdinfo;

/* Sets up and initializes the data structures */
int svd_allocate();

/* Finds the data structure associated w/a given movie or user */
inline svdinfo* svd_infoformovie(int movie_id);
inline svdinfo* svd_infoforuser(int user_id);

/* Deallocates the data structures */
int svd_cleanup();

/* Loads structures from file (or writes them to file) */
int svd_load(FILE *f);
int svd_save(FILE *f);

/* Does an iteration of ALS (or many) */
inline int svd_alsiterate(double lam, int *term);
void svd_alsrun(double lam, double cutoff, int maxiter, int *termslow, int *termfast);

/* Predicts a rating for a given movie&user */
inline double svd_predict(int mid, int uid);

/* Finds the error */
inline double svd_error(int visible);

/* Loads a (hypothetical or new) user from file */
rlist* svd_loaduser(FILE *f, double lam, double bias);

/* Makes n recommendations for a new user (whose known ratings are in rtmp) */
TopNNode* svd_recommend(rlist* rtmp, int n, double bias);

/* Finds the top (or bottom) movies along a particular dimension */
int svd_dimensions(int dim, int n);

#endif
