#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <math.h>

#include "../../netflixcore.h"
#include "svd.h"

/* Allocates the feature vectors and initialize them to values in (0.0,1.0) 
   (DO THIS AFTER HIDING THE TEST SET AND FINDING THE AVERAGE RATING) */
int svd_allocate() {
  rlist *rtmp;
  svdinfo *svdtmp, *svdtmp2;
  int i, j;
  double lam2, lam3;

  /* Regularization parameters for calculating the user and movie biases.
     Why these values? BelKor told me so! Same goes for the particular 
     method of approximating the bias terms for users and movies. */
  lam2 = 25; lam3 = 10;

  for(i = 0; i < NMOVIES; i++) {
    rtmp = nf_ratingsformovie(i);
    svdtmp = malloc(sizeof(svdinfo));
    rtmp->additional = svdtmp;
    if(svdtmp == NULL) {
      perror("Couldn't malloc SVD stuff");
      return 1;
    }
    svdtmp->err = -1;
    svdtmp->nvisible = 0;
    svdtmp->bias = 0.0;
    for(j = 0; j < rtmp->length; j++) {
      if(rtmp->ratings[j].visible) {
        svdtmp->nvisible++;
        svdtmp->bias += rtmp->ratings[j].stars - nf_ratings.ave;
      }
    }
    svdtmp->bias /= lam2 + svdtmp->nvisible;
    svdtmp->features = malloc(sizeof(feature)*NFEATURES);
    if(svdtmp->features == NULL) {
      perror("Couldn't malloc SVD stuff");
      return 1;
    }
    for(j = 0; j < NFEATURES; j++)
      svdtmp->features[j] = rand_0_1();
  }
  
  for(i = 0; i < NUSERS; i++) {
    rtmp = nf_ratingsforuser(i);
    svdtmp = malloc(sizeof(svdinfo));
    rtmp->additional = svdtmp;
    if(svdtmp == NULL) {
      perror("Couldn't malloc SVD stuff");
      return 1;
    }
    svdtmp->err = -1;
    svdtmp->nvisible = 0;
    svdtmp->bias = 0.0;
    for(j = 0; j < rtmp->length; j++) {
      if(rtmp->ratings[j].visible) {
        svdtmp2 = (svdinfo*)(nf_ratingsformovie(rtmp->ratings[j].id)->additional);
        svdtmp->nvisible++;
        svdtmp->bias += rtmp->ratings[j].stars - nf_ratings.ave - svdtmp2->bias;
      }
    }
    svdtmp->bias /= lam3 + svdtmp->nvisible;
    svdtmp->features = malloc(sizeof(feature)*NFEATURES);
    if(svdtmp->features == NULL) {
      perror("Couldn't malloc SVD stuff");
      return 1;
    }
    for(j = 0; j < NFEATURES; j++)
      svdtmp->features[j] = rand_0_1();
  }

  return 0;
}

/* Finds the svd info for a movie */
inline svdinfo* svd_infoformovie(int movie_id) {
  rlist* rtmp;
  rtmp = nf_ratingsformovie(movie_id);
  return (svdinfo*)rtmp->additional;
}

/* Finds the feature vector for a movie */
inline svdinfo* svd_infoforuser(int user_id) {
  rlist* rtmp;
  rtmp = nf_ratingsforuser(user_id);
  return (svdinfo*)rtmp->additional;
}

/* Deallocate memory associated with SVD */
int svd_cleanup() {
  rlist* rtmp;
  int i;
  for(i = 0; i < NUSERS; i++) {
    rtmp = nf_ratingsforuser(i);
    free(((svdinfo*)rtmp->additional)->features);
    free(rtmp->additional);
  }
  for(i = 0; i < NMOVIES; i++) {
    rtmp = nf_ratingsformovie(i);
    free(((svdinfo*)rtmp->additional)->features);
    free(rtmp->additional);
  }
  return 0;
}

/* Load feature vectors from file 
   NOTE: Call svd_cleanup first */
int svd_load(FILE *f) {
  int i, j, k;
  svdinfo *svdi;
  rlist *rtmp;

  /* Make sure file is OK */
  if(f == NULL) {
    perror("File error!\n");
    return 1;
  }

  /* Check that the precision matches */
  fread(&i, sizeof(int), 1, f);
  if(i != sizeof(feature)) {
    perror("Precision mismatch!\n");
    return 1;
  }

  /* Check that NMOVIES and NUSERS match */
  fread(&i, sizeof(int), 1, f);
  if(i != NMOVIES) {
    perror("NMOVIES mismatch!\n");
    return 1;
  }
  fread(&i, sizeof(int), 1, f);
  if(i != NUSERS) {
    perror("NUSERS mismatch!\n");
    return 1;
  }

  /* Find NFEATURES */
  fread(&j, sizeof(int), 1, f);
  if(j > NFEATURES) {
    perror("Too many features (in file)!");
    return 1;
  }

  /* Load the actual svdinfo structs */
  for(i = 0; i < NMOVIES; i++) {
    rtmp = nf_ratingsformovie(i);
    rtmp->additional = malloc(sizeof(svdinfo));
    svdi = (svdinfo*)(rtmp->additional);
    if(svdi == NULL) {
      perror("Couldn't malloc svdinfo");
      return 1;
    }
    svdi->features = malloc(sizeof(feature)*NFEATURES);
    if(svdi->features == NULL) {
      perror("Couldn't malloc features");
      return 1;
    }
    fread(svdi->features, sizeof(feature), j, f);
    for(k = j; k < NFEATURES; k++) {
      svdi->features[k] = 0.01;
    }
    fread(&(svdi->nvisible), sizeof(int), 1, f);
    fread(&(svdi->err), sizeof(double), 1, f);
    fread(&(svdi->bias), sizeof(double), 1, f);
  }
  for(i = 0; i < NUSERS; i++) {
    rtmp = nf_ratingsforuser(i);
    rtmp->additional = malloc(sizeof(svdinfo));
    svdi = (svdinfo*)(rtmp->additional);
    if(svdi == NULL) {
      perror("Couldn't malloc svdinfo");
      return 1;
    }
    svdi->features = malloc(sizeof(feature)*NFEATURES);
    if(svdi->features == NULL) {
      perror("Couldn't malloc features");
      return 1;
    }
    fread(svdi->features, sizeof(feature), j, f);
    for(k = j; k < NFEATURES; k++) {
      svdi->features[k] = 0.01;
    }
    fread(&(svdi->nvisible), sizeof(int), 1, f);
    fread(&(svdi->err), sizeof(double), 1, f);
    fread(&(svdi->bias), sizeof(double), 1, f);
  }

  fclose(f);
  return 0;
}

/* Save feature vectors to a file */
int svd_save(FILE *f) {
  int i, j;
  svdinfo *svdi;

  /* Make sure file is OK */
  if(f == NULL) {
    perror("File Error!\n");
    return 1;
  }

  /* Save the precision */
  i = sizeof(feature);
  fwrite(&i, sizeof(int), 1, f);

  /* Save NMOVIES and NUSERS */
  i = NMOVIES;
  fwrite(&i, sizeof(int), 1, f);
  i = NUSERS;
  fwrite(&i, sizeof(int), 1, f);

  /* Save NFEATURES */
  i = NFEATURES;
  fwrite(&i, sizeof(int), 1, f);

  /* Save the actual svdinfo structs */
  for(i = 0; i < NMOVIES; i++) {
    svdi = svd_infoformovie(i);
    fwrite(svdi->features, sizeof(feature), NFEATURES, f);
    fwrite(&(svdi->nvisible), sizeof(int), 1, f);
    fwrite(&(svdi->err), sizeof(double), 1, f);
    fwrite(&(svdi->bias), sizeof(double), 1, f);
  }
  for(i = 0; i < NUSERS; i++) {
    svdi = svd_infoforuser(i);
    fwrite(svdi->features, sizeof(feature), NFEATURES, f);
    fwrite(&(svdi->nvisible), sizeof(int), 1, f);
    fwrite(&(svdi->err), sizeof(double), 1, f);
    fwrite(&(svdi->bias), sizeof(double), 1, f);
  }

  fclose(f);
  return 0;
}

/* Error for user j using vec as the feature vector (length K)
   and ubias as the bias factor for the user */
double errforuser(int j, feature *vec, double ubias, int K) {
  double tmp, rmse, bias;
  int i, k;
  rlist *urlist;
  svdinfo *msvd;
  
  rmse = 0.0;
  bias = nf_ratings.ave;
  urlist = nf_ratingsforuser(j);
  for(i = 0; i < urlist->length; i++) {
    if(urlist->ratings[i].visible) {
      tmp = -1*urlist->ratings[i].stars;
      msvd = svd_infoformovie(urlist->ratings[i].id);
      tmp += bias + msvd->bias + ubias;
      for(k = 0; k < K; k++) {
        tmp += vec[k]*msvd->features[k];
      }
      tmp *= tmp;
      rmse += tmp;
    }
  }
  return rmse;
}

/* Error for movie i using vec as the feature vector (length K) and using
   mbias as the bias factor for the movie */
double errformovie(int i, feature *vec, double mbias, int K) {
  double tmp, rmse, bias;
  int j, k;
  rlist *mrlist;
  svdinfo *usvd;
  
  rmse = 0.0;
  mrlist = nf_ratingsformovie(i);
  bias = nf_ratings.ave;
  for(j = 0; j < mrlist->length; j++) {
    if(mrlist->ratings[j].visible) {
      tmp = -1*mrlist->ratings[j].stars;
      usvd = svd_infoforuser(mrlist->ratings[j].id);
      tmp += bias + usvd->bias + mbias;
      for(k = 0; k < K; k++) {
        tmp += vec[k]*usvd->features[k];
      }
      tmp *= tmp;
      rmse += tmp;
    }
  }
  return rmse;
}

/* Do an iteration of ALS */
inline int svd_alsiterate(double lam, int *term) {
  rlist *rtmp;
  int M, N, K, N_known, M_known, i, j, k, k2, itmp;
  svdinfo *svdtmp1, *svdtmp2;
  feature *KK, *Q_tmpi, *R_tmpj, *A_known, *R_known, *Q_known;
  double rmse, derr, tmp, bias;
  
  K = NFEATURES;
  M = NMOVIES;
  N = NUSERS;
  
  KK = malloc(K*K*sizeof(feature));
  Q_tmpi = malloc(K*sizeof(feature));
  R_tmpj = malloc(K*sizeof(feature));
  bias = nf_ratings.ave;  

  for(i = 0; i < M && !(*term); i++) {
    //printf("%d ", i);
    rtmp = nf_ratingsformovie(i);
    svdtmp1 = (svdinfo*)rtmp->additional;
    N_known = svdtmp1->nvisible;
    if(N_known == 0) continue;
    R_known = malloc(N_known*K*sizeof(feature));
    A_known = malloc(1*N_known*sizeof(feature));
    itmp = 0;
    for(j = 0; j < rtmp->length; j++) {
      if(rtmp->ratings[j].visible) {
        svdtmp2 = svd_infoforuser(rtmp->ratings[j].id);
        for(k = 0; k < K; k++) {
          R_known[itmp+k*N_known] = svdtmp2->features[k];
        }
        A_known[itmp] = rtmp->ratings[j].stars - bias 
                        - svdtmp1->bias - svdtmp2->bias;
        itmp++;
      }
    }
    
    for(k = 0; k < K; k++) {
      for(k2 = 0; k2 < K; k2++) {
        KK[k+k2*K] = 0.0;
        for(j = 0; j < N_known; j++) {
          KK[k+k2*K] += R_known[j+k*N_known]*R_known[j+k2*N_known];
        }
        if(k == k2) {
          KK[k+k2*K] += lam*N_known;
        }
      }
    }
    if(invmat(KK,K,K) != 0) {
      perror("Singular matrix :-(\n");
      free(Q_tmpi); free(R_tmpj); free(KK);
      free(R_known); free(A_known);
      return -1;
    }
    
    for(k = 0; k < K; k++) {
      Q_tmpi[k] = 0.0;
      for(j = 0; j < N_known; j++) {
        tmp = 0.0;
        for(k2 = 0; k2 < K; k2++) {
          tmp += KK[k+k2*K]*R_known[j+k2*N_known];
        }
        Q_tmpi[k] += tmp*A_known[j];
      }
    }
    
    free(R_known); free(A_known);
    
    /* Error for movie i, all its visible ratings, using Q_tmpi as the feature vector */
    rmse = errformovie(i, Q_tmpi, svdtmp1->bias, K);
    derr = svdtmp1->err - rmse;
    if(svdtmp1->err == -1) {
      derr = 1.0;
    }
        
    if(derr > 0.0 || 1) { /* or True? */
      memcpy(svdtmp1->features, Q_tmpi, K*sizeof(feature));
      svdtmp1->err = rmse;
    }
    else {
      printf("Increasing error @ Q-%d: %f to %f\n", i, svdtmp1->err, rmse);
      free(Q_tmpi); free(R_tmpj); free(KK);
      return -1;
    }
  }
  
  for(j = 0; j < N && !(*term); j++) {
    //printf("%d ", j);
    rtmp = nf_ratingsforuser(j);
    svdtmp1 = (svdinfo*)rtmp->additional;
    M_known = svdtmp1->nvisible;
    if(M_known == 0) continue;
    Q_known = malloc(M_known*K*sizeof(feature));
    A_known = malloc(1*M_known*sizeof(feature));
    itmp = 0;
    for(i = 0; i < rtmp->length; i++) {
      if(rtmp->ratings[i].visible) {
        svdtmp2 = svd_infoformovie(rtmp->ratings[i].id);
        for(k = 0; k < K; k++) {
          Q_known[itmp+k*M_known] = svdtmp2->features[k];
        }
        A_known[itmp] = rtmp->ratings[i].stars - bias
                        - svdtmp1->bias - svdtmp2->bias;
        itmp++;
      }
    }
    
    for(k = 0; k < K; k++) {
      for(k2 = 0; k2 < K; k2++) {
        KK[k+k2*K] = 0.0;
        for(i = 0; i < M_known; i++) {
          KK[k+k2*K] += Q_known[i+k*M_known]*Q_known[i+k2*M_known];
        }
        if(k == k2) {
          KK[k+k2*K] += lam*M_known;
        }
      }
    }
    
    if(invmat(KK,K,K) != 0) {
      perror("Singular matrix :-(\n");
      free(Q_tmpi); free(R_tmpj); free(KK);
      free(Q_known); free(A_known);
      return -1;
    }
    
    for(k = 0; k < K; k++) {
      R_tmpj[k] = 0.0;
      for(i = 0; i < M_known; i++) {
        tmp = 0.0;
        for(k2 = 0; k2 < K; k2++) {
          tmp += KK[k+k2*K]*Q_known[i+k2*M_known];
        }
        R_tmpj[k] += tmp*A_known[i];
      }
    }
    
    free(Q_known); free(A_known);
    
    /* Error for all movies rated by user j, using R_tmpj for a feature vector */
    rmse = errforuser(j, R_tmpj, svdtmp1->bias, K);
    derr = svdtmp1->err - rmse;
    if(svdtmp1->err == -1) {
      derr = 1.0;
    }
    
    /* printf("R %f\t%f\n", rmse, derr); */
    
    if(derr > 0.0 || 1) { /* or True? */
      memcpy(svdtmp1->features, R_tmpj, K*sizeof(feature));
      svdtmp1->err = rmse;
    }
    else {
      printf("Increasing error @ R-%d: %f to %f\n", j, svdtmp1->err, rmse);
      free(Q_tmpi); free(R_tmpj); free(KK);
      return -1;
    }
  }
  
  free(Q_tmpi); free(R_tmpj); free(KK);
  return 0;
}

/* Use the current feature vectors to predict a rating for a movie/user pair */
inline double svd_predict(int mid, int uid) {
  int i;
  double res, bias;
  svdinfo *mfeat, *ufeat;
  
  res = 0.0;
  bias = nf_ratings.ave;
  ufeat = svd_infoforuser(uid);
  mfeat = svd_infoformovie(mid);
  for(i = 0; i < NFEATURES; i++) {
    res += ufeat->features[i]*mfeat->features[i];
  }
  return res + bias + ufeat->bias + mfeat->bias;
}

/* Find the current error (total RMSE for visible/invisible ratings) */
inline double svd_error(int visible) {
  int i, j, k, n;
  double rmse, res, bias;
  svdinfo *mfeat, *ufeat;
  rlist *movie;

  bias = nf_ratings.ave;
  rmse = 0.0;
  n = 0;
  for(i = 0; i < NMOVIES; i++) {
    movie = nf_ratingsformovie(i);
    mfeat = ((svdinfo*)movie->additional);
    for(j = 0; j < movie->length; j++) {
      if(movie->ratings[j].visible == visible) {
        n++;
        ufeat = svd_infoforuser(movie->ratings[j].id);
        res = 0.0;
        for(k = 0; k < NFEATURES; k++) {
          res += ufeat->features[k]*mfeat->features[k];
        }
        res -= (double)movie->ratings[j].stars;
        res += bias + mfeat->bias + ufeat->bias;
        res *= res;
        rmse += res;
      }
    }
  }
  return sqrt(rmse/((double)n));
}

void svd_alsrun(double lam, double cutoff, int maxiter, int *termslow, int *termfast) {
  double derr, tse, lasttse, testrmse, bias;
  int i, j, k, nvisible, ninvisible, totstars, itmp;
  svdinfo *svdtmp;
  rlist *rtmp;
  time_t clock;
  struct tm *clockinfo;

  /* Count the number of visible and invisible ratings */
  nvisible = 0;
  ninvisible = 0;
  for(i = 0; i < NMOVIES; i++) {
    svdtmp = svd_infoformovie(i);
    rtmp = nf_ratingsformovie(i);
    nvisible += svdtmp->nvisible;
    ninvisible += rtmp->length;
  }
  ninvisible -= nvisible;
  bias = nf_ratings.ave;
  printf("%d ratings in training set\n", nvisible);
  printf("%d ratings in test set\n", ninvisible);
  printf("Ave rating: %f\n\n", bias);

  /* Do iterations of ALS-SVD, printing off progress stats */
  printf("trainerr\ttesterr\t\tderr (sq err)\ttime\n\n");
  for(i = 0; i < maxiter && !(*termslow); i++) {
    svd_alsiterate(lam, termfast);
    tse = 0.0;
    for(j = 0; j < NMOVIES; j++) {
      svdtmp = svd_infoformovie(j);
      if(svdtmp->nvisible > 0) {
        tse += svdtmp->err;
      }
    }
    testrmse = svd_error(0);
    if(i == 0) {
      derr = -1;
    }
    else {
      derr = (lasttse - tse)/nvisible;
    }
    time(&clock);
    clockinfo = localtime(&clock);
    printf("%f\t%f\t%f\t%s", sqrt(tse/nvisible), 
           testrmse, derr,
           asctime(clockinfo));
    lasttse = tse;
    if(derr >= 0.0 && derr <= cutoff) {
      printf("Done!\n");
      break;
    }
    else if(derr < 0.0 && i > 0) {
      printf("Worsening!\n");
      break;
    }
  }
}
