#include <assert.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "../../netflixcore.h"
#include "svd.h"

/* Thanks Jeff Ondich for the example signal handling code :-) */
void interrupt_handler(int sigid);
int term_slow = 0;
int term_fast = 0;
int mode;

#define FIND 0
#define USE  1
#define DIM  2
int main(int argc, char** argv) {
  /* Utility variables */
  time_t start, stop, loaded;
  struct tm *timeinfo;
  char *usage = "Usage:\nfsvd/dsvd -F [infile] outfile OR\nfsvd/dsvd -U infile ratingsfile N\nOR fsvd/dsvd -D infile DIM#(or -1) N\n";
  char *ifname, *ofname, *rfname, **titles;
  int n, dim, *mids;
  rlist *newuser;
  TopNNode *toplist, *p;

  /* Process command line args */
  if((argc == 3 || argc == 4) && strcmp("-F", argv[1]) == 0) {
    mode = FIND;
    rfname = NULL;
    if(argc == 3) {
      ifname = NULL;
      ofname = argv[2];
    }
    else {
      ifname = argv[2];
      ofname = argv[3];
    }
  }
  else if(argc == 5 && strcmp("-U", argv[1]) == 0) {
    mode = USE;
    ifname = argv[2];
    ofname = NULL;
    rfname = argv[3];
    n = atoi(argv[4]);
  }
  else if(argc == 5 && strcmp("-D", argv[1]) == 0) {
    mode = DIM;
    ifname = argv[2];
    ofname = NULL;
    rfname = NULL;
    dim = atoi(argv[3]);
    n = atoi(argv[4]);
  }
  else {
    fprintf(stderr, usage);
    exit(1);
  }

  /* Setup signal handling */
  if(signal(SIGINT, interrupt_handler) != SIG_DFL) {
    perror("Signal handling setup failed");
    exit(1);
  }

  /* Time it */
  time(&start);
  timeinfo = localtime(&start);
  printf("%s\n", asctime(timeinfo));

  /* Allocate the data structures */
  if(nf_allocate() != 0) {
    perror("Couldn't allocate Netflix stuff");
    exit(1);
  }

  if(mode == FIND) {
    /* Run over the training files, filling the data structures */
    if(nf_load() != 0) {
      perror("Couldn't load Netflix data");
      exit(1);
    } 

    /* Make the test set invisible */
    if(nf_hideprobeset() != 0) {
      perror("Couldn't hide test set");
      exit(1);
    }

    /* Calculate the average rating */
    if(nf_calcave() != 0) {
      perror("Couldn't calculate average rating");
      exit(1);
    }
  }
  
  /* Allocate and init the svd stuff (or load it) */
  if(ifname == NULL) {
    if(svd_allocate() != 0) {
      perror("Could allocate SVD stuff");
      exit(1);
    }
  }
  else {
    if(svd_load(fopen(ifname,"r")) != 0) {
      perror("Problem loading SVD vectors");
      exit(1);
    }
  }

  time(&loaded);
  timeinfo = localtime(&loaded);
  printf("Data loaded @ %s\n", asctime(timeinfo));

  /* Do the ALS-SVD */
  if(mode == FIND) {
    svd_alsrun(0.05, 0.001, 100, &term_slow, &term_fast);
  }
  /* Or make some recommendations */
  else if(mode == USE) {
    newuser = svd_loaduser(fopen(rfname, "r"), 0.01, 3.60);
    if(newuser == NULL) {
      perror("Failure loading user from file\n");
      exit(1);
    }
    toplist = svd_recommend(newuser, n, 3.60);
    if(toplist == NULL) {
      perror("Failure making recommendations");
      exit(1);
    }
    titles = nf_alloctitles(n);
    mids = topn_findoffsets(toplist, n);
    if(nf_findtitles(mids, titles, n) != 0) {
      perror("Failure finding titles");
      exit(1);
    }
    for(p = toplist; p != NULL; p = p->next) {
      printf("%f\t%d\t%s\n", p->rating, p->mid, titles[p->offset]);
    }
    free(mids); topn_freeall(toplist, n);
    nf_freetitles(titles, n);
  }
  else if(mode == DIM && svd_dimensions(dim, n) != 0) {
    perror("Failure analyzing dimensions");
    exit(1);
  }

  /* Save it */
  if(mode == FIND && svd_save(fopen(ofname,"w")) != 0) {
    perror("Problem saving SVD vectors");
    exit(1);
  }

  /* Clean up */
  svd_cleanup();
  nf_cleanup();

  /* Time it */
  time(&stop);
  timeinfo = localtime(&stop);
  printf("%s\n", asctime(timeinfo));
  printf("%f seconds elapsed\n", difftime(stop, start));

  return 0;
}

void interrupt_handler(int sigid) {
  int choice, flag;
  char *tmp;

  tmp = malloc(sizeof(char)*100);
  assert(tmp != NULL);

  if(mode == FIND) {
    flag = 0;
    fprintf(stderr, "\nDo you want to:\n1. Quit immediately?\n2. Quit fast (but save)?\n3. Quit after this iteration (and save)?\n4. Run to the end?\n >> ");
    while(!flag) {
      scanf("%100s", tmp);
      choice = atoi(tmp);
      if(choice >= 1 && choice <= 4) {
        break;
      }
      fprintf(stderr, "Try again\n >> ");
    }
    switch(choice) {
      case 1:
        exit(1);
      case 2:
        term_slow = term_fast = 1;
        break;
      case 3:
        term_slow = 1;
        term_fast = 0;
        break;
      case 4:
        term_slow = term_fast = 0;
        break;
    }
  }
  else {
    fprintf(stderr, "No signal handling for the current mode\n");
    exit(1);
  }

  free(tmp);
  fflush(stderr);
  (void)signal(SIGINT, interrupt_handler);
}
