'''svd.py

by Peter Nelson

Finds the best k-rank approximation to a matrix (in a least squares sense),
using SciPy's SVD implementation.

'''

from numpy import array, dot, zeros, empty, nonzero, identity
from numpy.linalg.linalg import LinAlgError
from numpy.linalg import inv
from scipy.linalg.decomp import svd, diagsvd
from scipy.linalg.basic import norm
from pylab import figure, plot, show, xlabel, ylabel, savefig, legend, title
import random

def infseq(start):
  '''Generate an infinite sequence beginning with start'''
  while True:
    yield start
    start += 1

def rand_matrix(m, n, choices):
  '''Generate an m x n matrix with entries drawn randomly from the choices'''
  result = zeros(shape=(m,n))
  for i in range(m):
    for j in range(n):
      result[i,j] = random.choice(choices)
  return result

def rand_mask(m, n, density):
  '''Generate an m x n binary matrix with ratio of 1's about density'''
  result = zeros(shape=(m,n))
  for i in range(m):
    for j in range(n):
      result[i,j] = int(random.random() < density)
  return result

def best_approx(matrix, k):
  '''Find the best k-rank approximation to a matrix'''
  A = array(matrix)
  M,N = A.shape
  U,Sig,Vt = svd(A)
  U = U
  Vt = Vt
  # Truncate values
  for i in range(k,len(Sig)):
    Sig[i] = 0
  Sig = diagsvd(Sig,M,N)
  return dot(dot(U,Sig),Vt)

def fdist(m1, m2, mask=None):
  '''Find the Frobenius distance between matrices'''
  assert m1.shape == m2.shape
  if mask is None:
    mask = zeros(shape=m1.shape)
    mask += 1
  else:
    assert m1.shape == mask.shape
  return norm(mask*(m1-m2), ord='fro')

def best_approx_gd1(matrix, K, lrate, derr_cutoff):
  '''Find the best k-rank approximation to a matrix using gradient descent'''
  A = array(matrix)
  M,N = A.shape
  Q = zeros(shape=(M,K))
  Q += .1
  R = zeros(shape=(K,N))
  R += .1
  old_lsq_err = norm(A, ord='fro')
  for k in range(K):
    while True:
      P = dot(Q,R)
      E = A - P
      new_lsq_err = norm(E, ord='fro')
      if (old_lsq_err - new_lsq_err) < derr_cutoff:
        break
      E *= lrate
      for m in range(M):
        for n in range(N):
          err = E[m,n]
          tmp = Q[m,k]
          Q[m,k] += err*R[k,n]
          R[k,n] += err*tmp
      old_lsq_err = new_lsq_err
  return dot(Q,R)

def best_approx_gd2(matrix, K, lrate, derr_cutoff):
  '''Find the best k-rank approximation to a matrix using gradient descent'''
  A = array(matrix)
  M,N = A.shape
  Q = zeros(shape=(M,K))
  Q += .1
  R = zeros(shape=(K,N))
  R += .1
  old_lsq_err = norm(A, ord='fro')
  while True:
    for m in range(M):
      for n in range(N):
        for k in range(K):
          P = dot(Q,R)
          E = A - P
          new_lsq_err = norm(E, ord='fro')
          if (old_lsq_err - new_lsq_err) < derr_cutoff:
            return P
          E *= lrate
          err = E[m,n]
          tmp = Q[m,k]
          Q[m,k] += err*R[k,n]
          R[k,n] += err*tmp
          old_lsq_err = new_lsq_err
  return dot(Q,R)

def best_approx_gd3(matrix, K, lrate, derr_cutoff):
  '''Find the best k-rank approximation to a matrix using gradient descent'''
  A = array(matrix)
  M,N = A.shape
  Q = zeros(shape=(M,K))
  Q += .1
  R = zeros(shape=(K,N))
  R += .1
  old_lsq_err = norm(A, ord='fro')
  while True:
    P = dot(Q,R)
    E = A - P
    new_lsq_err = norm(E, ord='fro')
    if (old_lsq_err - new_lsq_err) < derr_cutoff:
      break
    E *= lrate
    for m in range(M):
      for n in range(N):
        for k in range(K):
          err = E[m,n]
          tmp = Q[m,k]
          Q[m,k] += err*R[k,n]
          R[k,n] += err*tmp
    old_lsq_err = new_lsq_err
  return dot(Q,R)

def best_approx_als(matrix, K, derr_cutoff):
  '''Find the best k-rank approximation to a matrix using alternating least-squares'''
  A = array(matrix)
  M,N = A.shape
  Q = rand_matrix(M, K, range(1, 100))
  Q /= 100.0
  R = rand_matrix(K, N, range(1, 100))
  R /= 100.0
  old_lsq_err = norm(A, ord='fro')
  for iteration in infseq(1):
    P = dot(Q,R)
    E = A - P
    new_lsq_err = norm(E, ord='fro')
    derr = old_lsq_err - new_lsq_err
    if derr < derr_cutoff and iteration > 1:
      break
    RRt_invR = dot(inv(dot(R,R.T)),R)
    for i in range(M):
      Q[i,:] = dot(RRt_invR, A[i,:].T).T
    QtQ_invQt = dot(inv(dot(Q.T,Q)),Q.T)
    for j in range(N):
      R[:,j] = dot(QtQ_invQt, A[:,j].T)
    old_lsq_err = new_lsq_err
  #print iteration
  return dot(Q,R),iteration

def best_approx_als2(matrix, mask, K, lam, derr_cutoff):
  '''Find the best k-rank approximation to a matrix using alternating least-squares (w/missing values)'''
  A = array(matrix)
  M,N = A.shape
  Q = rand_matrix(M, K, range(1, 100))
  Q /= 100.0
  Q_tmp = Q.copy()
  R = rand_matrix(K, N, range(1, 100))
  R /= 100.0
  R_tmp = R.copy()
  errlog = [('INIT',norm(A*mask, ord = 'fro'))]
  for iteration in infseq(1):
    P = dot(Q,R)
    errlog.append(('P-%d' % iteration, fdist(P, A, mask)))
    if iteration > 1:
      derr = errlog[-4][1] - errlog[-1][1]
    else:
      derr = derr_cutoff + 1
    if 0 <= derr < derr_cutoff:
      #print iteration
      print errlog
      return P,iteration
    elif derr < 0:
      print "shit!"
      print errlog
      return P,iteration
    for i in range(M):
      known = nonzero(mask[i,:])[0]
      R_known = R[:, known]
      RRt = dot(R_known, R_known.T) + lam*len(known)*identity(K)
      RRt_invR = dot(inv(RRt),R_known)
      Q_tmp[i,:] = dot(RRt_invR, A[i,known].T).T
    errlog.append(('Q-%d-%d'%(iteration,i),fdist(dot(Q_tmp,R_tmp),A,mask)))
    if errlog[-1][1] < errlog[-2][1] or True:
      Q = Q_tmp.copy()
    else:
      print errlog
      print '(increasing error @ Q-%d-%d: %f to %f)' % (iteration,
                                                        i,
                                                        errlog[-2][1],
                                                        errlog[-1][1])
      raise
      Q_tmp = Q.copy()
      del errlog[-1]
    for j in range(N):
      known = nonzero(mask[:,j])[0]
      Q_known = Q[known,:]
      QtQ = dot(Q_known.T, Q_known) + lam*len(known)*identity(K)
      QtQ_invQt = dot(inv(QtQ),Q_known.T)
      R_tmp[:,j] = dot(QtQ_invQt, A[known,j].T)
    errlog.append(('R-%d-%d'%(iteration,j),fdist(dot(Q_tmp,R_tmp),A,mask)))
    if errlog[-1][1] < errlog[-2][1] or True:
      R = R_tmp.copy()
    else:
      print errlog
      print '(increasing error @ R-%d-%d: %f to %f)' % (iteration,
                                                        j,
                                                        errlog[-2][1],
                                                        errlog[-1][1])
      raise
      R_tmp = R.copy()
      del errlog[-1]

def best_approx_als3(matrix, mask, K, lam, derr_cutoff):
  '''Find the best k-rank approximation to a matrix using alternating least-squares (w/missing values)'''
  A = array(matrix)
  M,N = A.shape
  Q = rand_matrix(M, K, range(1, 100))
  Q /= 100.0
  Q_tmp = Q.copy()
  R = rand_matrix(N, K, range(1, 100))
  R /= 100.0
  R_tmp = R.copy()
  errlog = [('INIT',norm(A*mask, ord = 'fro'))]
  for iteration in infseq(1):
    P = dot(Q,R.T)
    errlog.append(('P-%d' % iteration, fdist(P, A, mask)))
    if iteration > 1:
      derr = errlog[-4][1] - errlog[-1][1]
    else:
      derr = derr_cutoff + 1
    if 0 <= derr < derr_cutoff:
      #print iteration
      #print errlog[-4:]
      return P,iteration
    elif derr < 0:
      print "doh!"
      #print errlog[-4:]
      print -1*derr
      return P,iteration
    for i in range(M):
      known = nonzero(mask[i,:])[0]
      R_known = R[known,:]
      RtR = dot(R_known.T, R_known) + lam*len(known)*identity(K)
      RtR_invRt = dot(inv(RtR),R_known.T)
      Q_tmp[i,:] = dot(RtR_invRt, A[i,known].T).T
    errlog.append(('Q-%d-%d'%(iteration,i),fdist(dot(Q_tmp,R_tmp.T),A,mask)))
    if errlog[-1][1] < errlog[-2][1] or True:
      Q = Q_tmp.copy()
    else:
      print errlog
      print '(increasing error @ Q-%d-%d: %f to %f)' % (iteration,
                                                        i,
                                                        errlog[-2][1],
                                                        errlog[-1][1])
      raise
      Q_tmp = Q.copy()
      del errlog[-1]
    for j in range(N):
      known = nonzero(mask[:,j])[0]
      Q_known = Q[known,:]
      QtQ = dot(Q_known.T, Q_known) + lam*len(known)*identity(K)
      QtQ_invQt = dot(inv(QtQ),Q_known.T)
      R_tmp[j,:] = dot(QtQ_invQt, A[known,j]).T
    errlog.append(('R-%d-%d'%(iteration,j),fdist(dot(Q_tmp,R_tmp.T),A,mask)))
    if errlog[-1][1] < errlog[-2][1] or True:
      R = R_tmp.copy()
    else:
      print errlog
      print '(increasing error @ R-%d-%d: %f to %f)' % (iteration,
                                                        j,
                                                        errlog[-2][1],
                                                        errlog[-1][1])
      raise
      R_tmp = R.copy()
      del errlog[-1]

def run_experiment(lim, density):
  A = rand_matrix(random.choice(range(2,lim)),
                  random.choice(range(2,lim)),
                  range(1,30))
  M,N = A.shape
  if density < max(M,N)/float(M*N):
    return None
  for trynum in infseq(1):
    mask = rand_mask(M, N, density)
    if all(sum(mask[i,:]) for i in range(M)) and all(sum(mask[:,j]) for j in range(N)):
      break
    elif trynum >= 100:
      return None
  A_0s = A.copy() * mask
  A_aves = A.copy() * mask
  for m in range(M):
    for n in range(N):
      if A_0s[m,n] == 0:
        A_aves[m,n] =  sum(A_0s[m,:])/sum(mask[m,:])
        A_aves[m,n] += sum(A_0s[:,n])/sum(mask[:,n])
        A_aves[m,n] /= 2.0
  lrate = 0.01
  derr_cutoff = 0.001
  res = {}
  res['derr_cutoff'] = derr_cutoff
  res['density'] = float(sum(sum(mask)))/(M*N)
  res['rank'] = min(M,N)
  res['ALS_rand'] = {'distance':[], 'iterations':[]}
  res['ALS_0'] = {'distance':[], 'iterations':[]}
  res['ALS_ave'] = {'distance':[], 'iterations':[]}
  res['ALS_0.02'] = {'distance':[], 'iterations':[]}
  res['ALS_0.04'] = {'distance':[], 'iterations':[]}
  res['ALS_0.06'] = {'distance':[], 'iterations':[]}
  for k in range(1, min(A.shape)+1):
    print '%d / %d' % (k, min(A.shape))
    tmpD,tmpI = best_approx_als(A, k, derr_cutoff)
    res['ALS_rand']['distance'].append(fdist(A, tmpD, mask))
    res['ALS_rand']['iterations'].append(tmpI)
    tmpD,tmpI = best_approx_als(A_0s, k, derr_cutoff)
    res['ALS_0']['distance'].append(fdist(A, tmpD, mask))
    res['ALS_0']['iterations'].append(tmpI)
    tmpD,tmpI = best_approx_als(A_aves, k, derr_cutoff)
    res['ALS_ave']['distance'].append(fdist(A, tmpD, mask))
    res['ALS_ave']['iterations'].append(tmpI)
    tmpD,tmpI = best_approx_als3(A, mask, k, 0.02, derr_cutoff)
    res['ALS_0.02']['distance'].append(fdist(A, tmpD, mask))
    res['ALS_0.02']['iterations'].append(tmpI)
    tmpD,tmpI = best_approx_als3(A, mask, k, 0.04, derr_cutoff)
    res['ALS_0.04']['distance'].append(fdist(A, tmpD, mask))
    res['ALS_0.04']['iterations'].append(tmpI)
    tmpD,tmpI = best_approx_als3(A, mask, k, 0.06, derr_cutoff)
    res['ALS_0.06']['distance'].append(fdist(A, tmpD, mask))
    res['ALS_0.06']['iterations'].append(tmpI)
  return res

for expnum in range(100):
  results = None
  while not results:
    results = run_experiment(40, random.random()/5.0)
  figure()
  title('Low-Rank Approximations to R (density %f)' % results['density'])
  xlabel('num features')
  ylabel('error')
  xlist = range(1,results['rank']+1)
  for alg,col in zip(['ALS_rand', 'ALS_0', 'ALS_ave', 'ALS_0.02', 'ALS_0.04', 'ALS_0.06'],
                     ['r', 'g', 'b', 'c', 'm', 'y']):
    ylist = results[alg]['distance']
    assert len(xlist) == len(ylist), "mismatch %d %d" % (len(xlist), len(ylist))
    plot(xlist, ylist, col+'-', label=alg)
  legend()
  savefig("graphs/err_%d.png" % expnum)
  #raw_input('...')
  figure()
  title('Low-Rank Approximations to R (density %f)' % results['density'])
  xlabel('num features')
  ylabel('iterations to convergence')
  xlist = range(1,results['rank']+1)
  for alg,col in zip(['ALS_rand', 'ALS_0', 'ALS_ave', 'ALS_0.02', 'ALS_0.04', 'ALS_0.06'],
                     ['r', 'g', 'b', 'c', 'm', 'y']):
    ylist = results[alg]['iterations']
    assert len(xlist) == len(ylist), "mismatch %d %d" % (len(xlist), len(ylist))
    plot(xlist, ylist, col+'-', label=alg)
  legend()
  savefig("graphs/iter_%d.png" % expnum)

